# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .bulk import BulkOperation
from .document import Document, SecurityDocument
from .design_document import DesignDocument
from .exception import ConflictError, NotFoundError
from .remote import RemoteDatabase
from .view import AllDocsView, View


class Database(RemoteDatabase):
    def __init__(self, couchdb, id):
        super().__init__(couchdb._server, id)

    async def akeys(self, **params):
        async for key in self.all_docs.ids(**params):
            yield key

    async def create(self, id, exists_ok=False):
        doc = Document(self, id)

        if exists_ok:
            try:
                await doc.fetch(discard_changes=True)
            except NotFoundError:
                pass
        else:
            if await doc._exists():
                raise ConflictError(
                    f"The document '{id}' does already exist in the database '{self.id}'"
                )

        return doc

    async def delete(self):
        await self._delete()

    async def docs(self, ids=None, create=False, prefix=None, **params):
        async for doc in self.all_docs.docs(ids, create, prefix, **params):
            yield doc

    async def values(self, **params):
        async for doc in self.all_docs.docs(**params):
            yield doc

    @property
    def all_docs(self):
        return AllDocsView(self)

    def view(self, design_doc, view):
        return View(self, design_doc, view)

    async def design_doc(self, id, exists_ok=False):
        ddoc = DesignDocument(self, id)

        if await ddoc._exists():
            if exists_ok:
                await ddoc.fetch(discard_changes=True)
            else:
                raise ConflictError(
                    f"The design document '{id}' does already exist in the database '{self.id}'"
                )

        return ddoc

    async def find(self, selector, limit=None, **params):
        # we need to get the complete doc, so fields selector isn't allowed
        if "fields" in selector.keys():
            raise ValueError("selector must not contain a fields entry")

        # We must use pagination because otherwise the default limit of the _find endpoint
        # fucks us
        pagination_size = limit if limit is not None else 10000

        while True:
            result_chunk = await self._find(selector, limit=pagination_size, **params)

            for res in result_chunk["docs"]:
                doc = Document(self, res["_id"])
                doc._update_cache(res)
                yield doc

            if len(result_chunk["docs"]) < pagination_size or limit is not None:
                break

            params["bookmark"] = result_chunk["bookmark"]

    def update_docs(self, ids, create=False):
        return BulkOperation(self, ids, create)

    async def __getitem__(self, id):
        return await self.get(id)

    async def get(self, id, default=None):
        doc = Document(self, id)

        try:
            await doc.fetch(discard_changes=True)
        except NotFoundError as e:
            if default is not None:
                doc.update(default)
            else:
                raise e

        return doc

    async def security(self):
        doc = SecurityDocument(self)
        await doc.fetch(discard_changes=True)
        return doc

    async def info(self):
        return await self._get()



import aiohttp

from contextlib import suppress
import functools


class BadRequestError(ValueError):
    pass


class UnauthorizedError(PermissionError):
    pass


class ForbiddenError(PermissionError):
    pass


class NotFoundError(KeyError):
    pass


class ConflictError(KeyError):
    pass


class ExpectationFailedError(ValueError):
    pass


class PreconditionFailedError(RuntimeError):
    pass


class UnsupportedMediaTypeError(ValueError):
    pass


def raise_for_endpoint(endpoint, message, exception, exception_type=None):
    if exception_type is None:
        if exception.status == 400:
            exception_type = BadRequestError
        elif exception.status == 401:
            exception_type = UnauthorizedError
        elif exception.status == 403:
            exception_type = ForbiddenError
        elif exception.status == 404:
            exception_type = NotFoundError
        elif exception.status == 409:
            exception_type = ConflictError
        elif exception.status == 412:
            exception_type = PreconditionFailedError
        elif exception.status == 415:
            exception_type = UnsupportedMediaTypeError
        elif exception.status == 417:
            exception_type = ExpectationFailedError
        else:
            raise ValueError(
                "Something went wrong, but I couldn't deduce the type of exception nor "
                "format a nice exception for you. I'm sorry."
            ) from exception

    message_input = {}

    with suppress(AttributeError):
        message_input["id"] = endpoint.id
        message_input["endpoint"] = endpoint.endpoint
        message_input["rev"] = endpoint._data.get("_rev")
    with suppress(AttributeError):
        message_input["document_id"] = endpoint._document.id
        message_input["document_rev"] = endpoint._document._data.get("_rev")

    raise exception_type(message.format(**message_input)) from exception


def raises(status, message, exception_type=None):
    def decorator_raises(func):
        @functools.wraps(func)
        async def wrapper(endpoint, *args, **kwargs):
            try:
                return await func(endpoint, *args, **kwargs)
            except aiohttp.ClientResponseError as exception:
                if status == exception.status:
                    raise_for_endpoint(endpoint, message, exception, exception_type)
                raise exception

        return wrapper

    return decorator_raises



from .couchdb import CouchDB
from .exception import (
    BadRequestError,
    ConflictError,
    UnauthorizedError,
    ForbiddenError,
    NotFoundError,
    ExpectationFailedError,
    PreconditionFailedError,
    UnsupportedMediaTypeError,
)

__all__ = [
    "CouchDB",
    "BadRequestError",
    "ConflictError",
    "UnauthorizedError",
    "ForbiddenError",
    "NotFoundError",
    "ExpectationFailedError",
    "PreconditionFailedError",
    "UnsupportedMediaTypeError",
]



from .exception import ConflictError, NotFoundError
from .database import Database
from .remote import RemoteServer


class CouchDB(object):
    def __init__(self, *args, **kwargs):
        self._server = RemoteServer(*args, **kwargs)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    async def check_credentials(self):
        await self._server._check_session()

    async def close(self):
        await self._server.close()

    async def create(self, id, exists_ok=False, **kwargs):
        db = Database(self, id)
        if not await db._exists():
            await db._put(**kwargs)
            return db
        elif exists_ok:
            return db
        else:
            raise ConflictError(f"The database '{id}' does already exist.")

    async def __getitem__(self, id):
        db = Database(self, id)

        if not await db._exists():
            raise NotFoundError(f"The database '{id}' does not exist.")

        return db

    async def keys(self, **params):
        return await self._server._all_dbs(**params)

    async def info(self):
        return await self._server._info()



from setuptools import setup

test_deps = ["pytest", "pytest-asyncio", "pytest-cov", "packaging"]

setup(
    name="aiocouch",
    version="1.0.1",
    license="BSD 3-clause",
    description="🛋 An asynchronous client library for CouchDB 2.x",
    author="TU Dresden",
    url="https://github.com/metricq/aiocouch",
    keywords=["asyncio", "couchdb"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.6",
    packages=["aiocouch"],
    scripts=[],
    install_requires=["aiohttp~=3.0"],
    setup_requires=["pytest-runner"],
    tests_require=test_deps,
    extras_require={
        "examples": ["aiomonitor", "click", "click-log", "click-completion"],
        "tests": test_deps,
    },
)



from .document import Document
from .exception import ConflictError
from .remote import _quote_id
from .view import View


class DesignDocument(Document):
    _allowed_keys = [
        "language",
        "options",
        "filters",
        "lists",
        "rewrites",
        "shows",
        "updates",
        "validate_doc_update",
        "views",
    ]

    @property
    def endpoint(self):
        return f"{self._database.endpoint}/_design/{_quote_id(self.id)}"

    def __setitem__(self, key, value):
        if key in self._allowed_keys:
            super().__setitem__(key, value)
        else:
            raise KeyError(f"The key '{key}' is not allowed in an design document.")

    def view(self, view):
        return View(self._database, self.id, view)

    async def create_view(
        self, view, map_function, reduce_function=None, exists_ok=False
    ):
        if "views" not in self:
            self["views"] = {}

        if view in self["views"] and not exists_ok:
            raise ConflictError(
                f"The view '{view}' does already exist in the design document {self.id}"
            )

        self["views"][view] = {"map": map_function}
        if reduce_function is not None:
            self["views"][view]["reduce"] = reduce_function
        self["language"] = "javascript"

        await self.save()

        return self.view(view)
